import numpy as np
import datetime as dt
import spaceweather as sw
import pandas as pd

from .utils import get_correctly_formatted_particle_dist_list
from .anisotropic_MAIRE_engine.spectralCalculations.rigiditySpectrum import DLRmodelSpectrum, CommonModifiedPowerLawSpectrum, CommonModifiedPowerLawSpectrumSplit
from .anisotropic_MAIRE_engine.spectralCalculations.pitchAngleDistribution import gaussianBeeckPitchAngleDistribution, isotropicPitchAngleDistribution, gaussianPitchAngleDistribution
from .anisotropic_MAIRE_engine.generalEngineInstance import generalEngineInstance, default_array_of_lats_and_longs

default_altitudes_in_kft = [0,10,20] + [i for i in range(25, 61 + 1, 3)]

def run_from_spectra(
        proton_rigidity_spectrum=None,
        alpha_rigidity_spectrum=None,
        reference_pitch_angle_latitude=0.0, #None,
        reference_pitch_angle_longitude=45.0, #None,
        proton_pitch_angle_distribution=isotropicPitchAngleDistribution(),
        alpha_pitch_angle_distribution=isotropicPitchAngleDistribution(),
        altitudes_in_kft=default_altitudes_in_kft,
        altitudes_in_km=None,
        Kp_index=None,
        date_and_time=dt.datetime.utcnow(),
        array_of_lats_and_longs=default_array_of_lats_and_longs,
        cache_magnetocosmics_run=True,
        generate_NM_count_rates=False,
        use_default_9_zeniths_azimuths=False,
        **mag_cos_kwargs,
):
    
    if Kp_index is None:
        Kp_index = sw.ap_kp_3h()[pd.to_datetime(sw.ap_kp_3h().index) < date_and_time].iloc[-1]["Kp"]
        #raise Exception("Error: no Kp index specified!")
    
    if (altitudes_in_km is not None) and (altitudes_in_kft is not default_altitudes_in_kft):
        raise Exception("Error: only one of altitudes_in_km and altitudes_in_kft should be supplied!")
    
    if altitudes_in_km is None:
        altitudes_in_km = np.array(altitudes_in_kft) * 0.3048

    if (proton_rigidity_spectrum is None) and (alpha_rigidity_spectrum is None):
        raise Exception("Error: either a proton rigidity spectrum or an alpha rigidity spectrum must be specified!")
    
    list_of_particle_distributions = get_correctly_formatted_particle_dist_list(proton_rigidity_spectrum, 
                                                                                alpha_rigidity_spectrum, 
                                                                                reference_pitch_angle_latitude, 
                                                                                reference_pitch_angle_longitude, 
                                                                                proton_pitch_angle_distribution, 
                                                                                alpha_pitch_angle_distribution)

    
    engine_to_run = generalEngineInstance(list_of_particle_distributions,
                                          list_of_altitudes_km=altitudes_in_km,
                                          Kp_index=Kp_index,
                                          date_and_time=date_and_time,
                                          reference_latitude=reference_pitch_angle_latitude,
                                          reference_longitude=reference_pitch_angle_longitude,
                                          array_of_lats_and_longs=array_of_lats_and_longs,
                                          cache_magnetocosmics_runs=cache_magnetocosmics_run,
                                          generate_NM_count_rates=generate_NM_count_rates)
    
    output_dose_rate_DF = engine_to_run.getAsymptoticDirsAndRun(use_default_9_zeniths_azimuths,**mag_cos_kwargs)

    print("Success!")

    return output_dose_rate_DF

def run_from_power_law_gaussian_distribution(J0, gamma, deltaGamma, sigma, 
                                             reference_pitch_angle_latitude, reference_pitch_angle_longitude, 
                                             Kp_index,date_and_time,
                                             use_split_spectrum=False,
                                             **kwargs):
    
    if use_split_spectrum == True:
        spec_to_use = CommonModifiedPowerLawSpectrumSplit
    else:
        spec_to_use = CommonModifiedPowerLawSpectrum
        

    output_dose_rate_DF = run_from_spectra(
        proton_rigidity_spectrum=spec_to_use(J0, gamma, deltaGamma),
        reference_pitch_angle_latitude=reference_pitch_angle_latitude,
        reference_pitch_angle_longitude=reference_pitch_angle_longitude,
        proton_pitch_angle_distribution=gaussianPitchAngleDistribution(normFactor=1,sigma=sigma),
        Kp_index=Kp_index,date_and_time=date_and_time,
        **kwargs,
    )

    return output_dose_rate_DF

def run_from_double_power_law_gaussian_distribution(J0, gamma, deltaGamma, sigma_1, sigma_2,
                                                    B, alpha_prime,
                                             reference_pitch_angle_latitude, reference_pitch_angle_longitude, 
                                             Kp_index,date_and_time,
                                             use_split_spectrum=False,
                                             **kwargs):
    
    if use_split_spectrum == True:
        spec_to_use = CommonModifiedPowerLawSpectrumSplit
    else:
        spec_to_use = lambda J0,gamma,deltaGamma:CommonModifiedPowerLawSpectrum(J0,gamma,deltaGamma, lowerLimit=0.814529,upperLimit=21.084584)
        

    output_dose_rate_DF = run_from_spectra(
        proton_rigidity_spectrum=spec_to_use(J0, gamma, deltaGamma),
        reference_pitch_angle_latitude=reference_pitch_angle_latitude,
        reference_pitch_angle_longitude=reference_pitch_angle_longitude,
        proton_pitch_angle_distribution=gaussianPitchAngleDistribution(normFactor=1,sigma=sigma_1) + (B * gaussianPitchAngleDistribution(normFactor=1,sigma=sigma_2,alpha=alpha_prime)),
        Kp_index=Kp_index,date_and_time=date_and_time,
        **kwargs,
    )

    return output_dose_rate_DF

def run_from_power_law_Beeck_gaussian_distribution(J0, gamma, deltaGamma, A, B, 
                                             reference_pitch_angle_latitude, reference_pitch_angle_longitude, 
                                             Kp_index,date_and_time,
                                             use_split_spectrum=False,
                                             **kwargs):
    
    if use_split_spectrum == True:
        spec_to_use = CommonModifiedPowerLawSpectrum
    else:
        spec_to_use = CommonModifiedPowerLawSpectrumSplit

    output_dose_rate_DF = run_from_spectra(
        proton_rigidity_spectrum=spec_to_use(J0, gamma, deltaGamma),
        reference_pitch_angle_latitude=reference_pitch_angle_latitude,
        reference_pitch_angle_longitude=reference_pitch_angle_longitude,
        proton_pitch_angle_distribution=gaussianBeeckPitchAngleDistribution(normFactor=1,A=A,B=B),
        Kp_index=Kp_index,date_and_time=date_and_time,
        **kwargs,
    )

    return output_dose_rate_DF

def run_from_DLR_cosmic_ray_model(OULU_count_rate_in_seconds=None,
                                      W_parameter=None,
                                      Kp_index=None,date_and_time=None,
                                      **kwargs):
    
    if (W_parameter is None) and (OULU_count_rate_in_seconds is None):
        print("As no OULU count rates or W parameters were specified, the count rate of OULU in the past will be determined using the supplied date and time for the purposes of calculating the incoming spectra.")
        DLR_model_date_and_time = date_and_time
    else:
        DLR_model_date_and_time = None

    output_dose_rate_DF = run_from_spectra(
        proton_rigidity_spectrum=DLRmodelSpectrum(atomicNumber=1, date_and_time=DLR_model_date_and_time, OULUcountRateInSeconds=OULU_count_rate_in_seconds, W_parameter=W_parameter),
        alpha_rigidity_spectrum=DLRmodelSpectrum(atomicNumber=2, date_and_time=DLR_model_date_and_time, OULUcountRateInSeconds=OULU_count_rate_in_seconds, W_parameter=W_parameter),
        Kp_index=Kp_index,date_and_time=date_and_time,
        **kwargs,
    )

    return output_dose_rate_DF




    



import numpy as np
import pandas as pd
import datetime as dt
import sys
import ParticleRigidityCalculationTools as PRCT
from joblib import Memory
import tqdm
tqdm.tqdm.pandas()
from pandarallel import pandarallel
pandarallel.initialize(progress_bar=True)
from spacepy.coordinates import Coords as spaceCoords
from spacepy.time import Ticktock as spaceTicktock
import numba

from .spectralCalculations.particleDistribution import particleDistribution

memory_asymp_dirs = Memory("./cacheAsymptoticDirectionOutputs", verbose=0)

m0 = 1.67262192e-27 #kg
c = 299792458.0 #m/s
protonCharge = 1.60217663e-19 #C
protonRestEnergy = m0 * (c**2)

global get_apply_method
def get_apply_method(DF_or_Series):
    if hasattr(sys, 'gettrace') and sys.gettrace() is not None:
        print("debug mode being used: setting AniMAIRE to use progress_apply rather than running in parallel!")
        apply_method = DF_or_Series.progress_apply
    else:
        #print("not in debug mode: setting AniMAIRE to use parallel_apply!")
        apply_method = DF_or_Series.parallel_apply
        
    return apply_method

def generate_asymp_dir_DF(dataframeToFillFrom:pd.DataFrame, IMFlatitude:float, IMFlongitude:float, datetime_to_run_across_UTC, cache:bool):

    new_asymp_dir_DF = dataframeToFillFrom.copy()

    #new_asymp_dir_DF["Energy"] = PRCT.convertParticleRigidityToEnergy(dataframeToFillFrom["Rigidity"], particleMassAU = 1, particleChargeAU = 1)

    print("assigning asymptotic coordinates")
    if cache == False:
        asymptoticDirectionList = convertAsymptoticDirectionsToPitchAngle(dataframeToFillFrom, IMFlatitude, IMFlongitude, datetime_to_run_across_UTC)
    else:
        cachedConvertAsympDirFunc = memory_asymp_dirs.cache(convertAsymptoticDirectionsToPitchAngle)
        asymptoticDirectionList = cachedConvertAsympDirFunc(dataframeToFillFrom, IMFlatitude, IMFlongitude, datetime_to_run_across_UTC)

    print("successfully converted asymptotic directions")

    new_asymp_dir_DF["angleBetweenIMFinRadians"] = asymptoticDirectionList

    return new_asymp_dir_DF

def convertAsymptoticDirectionsToPitchAngle(dataframeToFillFrom:pd.DataFrame, IMFlatitude:float, IMFlongitude:float, datetime_to_run_across_UTC:dt.datetime):

    print("acquiring pitch angles...")
    pitch_angle_list = get_apply_method(dataframeToFillFrom)(lambda dataframe_row:get_pitch_angle_for_DF_analytic(IMFlatitude, IMFlongitude, dataframe_row["Lat"], dataframe_row["Long"]),axis=1)

    return pitch_angle_list

@numba.jit(nopython=True)
def get_pitch_angle_for_DF_analytic(IMFlatitude:float, IMFlongitude:float, asymptotic_dir_latitude:float, asymptotic_dir_longitude:float):

    IMFlatitude_rad = IMFlatitude * (np.pi/180.0)
    IMFlongitude_rad = IMFlongitude * (np.pi/180.0)
    asymptotic_dir_latitude_rad = asymptotic_dir_latitude * (np.pi/180.0)
    asymptotic_dir_longitude_rad = asymptotic_dir_longitude * (np.pi/180.0)

    cos_pitch_angle = (np.sin(asymptotic_dir_latitude_rad) * np.sin(IMFlatitude_rad)) + \
                      (np.cos(asymptotic_dir_latitude_rad) * np.cos(IMFlatitude_rad) * np.cos(asymptotic_dir_longitude_rad - IMFlongitude_rad))
    
    pitch_angle = np.arccos(cos_pitch_angle)

    return pitch_angle

def acquireWeightingFactors(asymptotic_direction_DF:pd.DataFrame, particle_dist:particleDistribution):

    momentaDist = particle_dist.momentum_distribution
    new_asymptotic_direction_DF = asymptotic_direction_DF.copy()

    # find weighting factors from the angles and rigidities
    #jacobian_function_to_use = lambda pitch_angle_in_radians:1/np.sin(2.0 * pitch_angle_in_radians)
        
    pitchAngleFunctionToUse = lambda row : momentaDist.getPitchAngleDistribution()(row["angleBetweenIMFinRadians"],row["Rigidity"])
    fullRigidityPitchWeightingFactorFunctionToUse = lambda row : momentaDist(row["angleBetweenIMFinRadians"],row["Rigidity"])

    print("calculating pitch angle weighting factors...")
    new_asymptotic_direction_DF["PitchAngleWeightingFactor"] = get_apply_method(new_asymptotic_direction_DF)(pitchAngleFunctionToUse,axis=1)

    new_asymptotic_direction_DF["Filter"] = (new_asymptotic_direction_DF["Filter"] == 1) * 1

    print("calculating rigidity weighting factors...")
    new_asymptotic_direction_DF["RigidityWeightingFactor"] = get_apply_method(new_asymptotic_direction_DF["Rigidity"])(momentaDist.getRigiditySpectrum())

    print("calculating rigidity + pitch combined weighting factors...")
    all_weighted_asymp_dirs = get_apply_method(new_asymptotic_direction_DF)(fullRigidityPitchWeightingFactorFunctionToUse,axis=1)
    new_asymptotic_direction_DF["fullRigidityPitchWeightingFactor"] = all_weighted_asymp_dirs * (new_asymptotic_direction_DF["Filter"] == 1)
    
    print("calculating energy + pitch combined weighting factors...")
    new_asymptotic_direction_DF["Energy"] = PRCT.convertParticleRigidityToEnergy(new_asymptotic_direction_DF["Rigidity"], 
                                                                      particleMassAU = particle_dist.particle_species.atomicMass, 
                                                                      particleChargeAU = particle_dist.particle_species.atomicNumber)
    energySpectrum = PRCT.convertParticleRigiditySpecToEnergySpec(new_asymptotic_direction_DF["Rigidity"],
                                                                  new_asymptotic_direction_DF["fullRigidityPitchWeightingFactor"],
                                                                  particleMassAU=particle_dist.particle_species.atomicMass, 
                                                                  particleChargeAU=particle_dist.particle_species.atomicNumber)

    new_asymptotic_direction_DF["fullEnergyPitchWeightingFactor"] = energySpectrum["Energy distribution values"]

    return new_asymptotic_direction_DF

def calculatePitchAngle_from_IMF_dir(interplanetary_mag_field:spaceCoords, asymptotic_direction:spaceCoords, datetime_to_run_across_UTC):

    cartesianAsympDir = asymptotic_direction.convert("GEO","car").data[0]
    GEOIMF = interplanetary_mag_field
    GEOIMF.ticks = spaceTicktock(datetime_to_run_across_UTC,"UTC")
    cartesianIMF = GEOIMF.convert("GEO","car").data[0]
    return calculateAngleBetweenTheSpaceVectors(cartesianAsympDir, cartesianIMF)

def calculatePitchAngle(momentaDist, dfRow, datetime_to_run_across_UTC):

    cartesianAsympDir = dfRow["Asymptotic Direction"].convert("GEO","car").data[0]
    GEOIMF = momentaDist.getPitchAngleDistribution().interplanetary_mag_field
    GEOIMF.ticks = spaceTicktock(datetime_to_run_across_UTC,"UTC")
    cartesianIMF = GEOIMF.convert("GEO","car").data[0]
    return calculateAngleBetweenTheSpaceVectors(cartesianAsympDir, cartesianIMF)

def calculateAngleBetweenTheSpaceVectors(cartesianAsympDir, cartesianIMF):
    dotProduct = np.dot(cartesianAsympDir, cartesianIMF)
    amplitude = np.linalg.norm(cartesianAsympDir)*np.linalg.norm(cartesianIMF)
    angleBetweenIMF = np.arccos(dotProduct/amplitude)
    return angleBetweenIMF



import numpy as np
import plotly.express as px
from scipy.interpolate import NearestNDInterpolator
import matplotlib.pyplot as plt
import seaborn as sns
import geopandas

def plot_dose_map_contours(dose_map_to_plot,levels=3,**kwargs):

    dose_map_to_plot_sorted = dose_map_to_plot.sort_values(by=["longitudeTranslated","latitude"])

    (contour_longs, contour_lats) = np.meshgrid(dose_map_to_plot_sorted["longitudeTranslated"].unique(),
            dose_map_to_plot_sorted["latitude"].unique())

    interp = NearestNDInterpolator(list(zip(dose_map_to_plot_sorted["longitudeTranslated"], dose_map_to_plot_sorted["latitude"])),
                               dose_map_to_plot_sorted["edose"])

    contours = plt.contour(contour_longs,contour_lats,interp(contour_longs, contour_lats),
            levels=levels,linestyles="dashed",colors="black",zorder=1000,**kwargs)
    plt.clabel(contours, inline=True) #,fmt={"fontweight":"bold"})

def create_single_dose_map_plot_plt(heatmap_DF_to_Plot,
                                    hue_range = None, #(0,100), 
                                    heatmap_s = 63,
                                    edgecolor=None,
                                    dose_type = "edose",
                                    legend_label=r"Effective dose ($\mu Sv / hr$)",
                                    palette="Spectral_r",
                                    plot_longitude_east=False,
                                    plot_colorbar=True):

    if not (heatmap_DF_to_Plot["altitude (km)"].nunique() == 1):
        print()
        print("\033[1mWARNING: multiple altitudes were supplied in the input dataframe, therefore only the map for the maximum altitude will be plotted!\033[0m")
        print()
        heatmap_DF_to_Plot = heatmap_DF_to_Plot.query(f"`altitude (km)` == {heatmap_DF_to_Plot['altitude (km)'].max()}")

    if hue_range is None:
        hue_range = (0,heatmap_DF_to_Plot["edose"].max())

    ############################ creating background world map and dose image
    currentFigure = plt.gcf()

    currentFigure.set_figheight(10)
    currentFigure.set_figwidth(10)

    #heatmap_DF_to_Plot = pd.read_csv(file_path_to_read, delimiter=',')
    heatmap_DF_to_Plot["SEU (Upsets/hr/Gb)"] = heatmap_DF_to_Plot["SEU"] * (60.0 * 60.0) * 1e9
    heatmap_DF_to_Plot["SEL (Latch-ups/hr/device)"] = heatmap_DF_to_Plot["SEL"] * (60.0 * 60.0)
    if plot_longitude_east is False:
        heatmap_DF_to_Plot["longitudeTranslated"] = heatmap_DF_to_Plot["longitude"].apply(lambda x:x-360.0 if x > 180.0 else x)
    else:
        heatmap_DF_to_Plot["longitudeTranslated"] = heatmap_DF_to_Plot["longitude"]

    scatterPlotAxis = sns.scatterplot(data=heatmap_DF_to_Plot,x="longitudeTranslated",y="latitude",
                    hue=dose_type, hue_norm=hue_range, palette=palette,
                    zorder=10,
                    marker="s",s=heatmap_s,edgecolor=edgecolor,
                    legend=False,
                    )#ax=axToPlotOn)

    if plot_colorbar is True:
        norm = plt.Normalize(hue_range[0], hue_range[1])
        sm = plt.cm.ScalarMappable(cmap=palette, norm=norm)
        sm.set_array([])

        # Remove the legend and add a colorbar
        #scatterPlotAxis.get_legend().remove()
        #colorbar = scatterPlotAxis.figure.colorbar(sm,label=legend_label,shrink=0.4)
        colorbar = scatterPlotAxis.figure.colorbar(sm,label=legend_label,orientation="horizontal")
    else:
        colorbar = None

    plt.ylim([-90,90])
    plt.xlim([-175,180])
    plt.grid(True)
    plt.xlabel("Longitude (degrees)")
    plt.ylabel("Latitude (degrees)")

    world = geopandas.read_file(geopandas.datasets.get_path('naturalearth_lowres'))
    world.plot(color="None",edgecolor="black",lw=0.35,ax=scatterPlotAxis,zorder=20)
    if plot_longitude_east is True:
        world['geometry'] = world['geometry'].translate(xoff=360)
        world.plot(color="None",edgecolor="black",lw=0.35,ax=scatterPlotAxis,zorder=20)
        plt.xlim([0,355])

    ####################################################################

    #plt.legend(title=dose_type,loc="center left",bbox_to_anchor=(1.1,0.5))

    return scatterPlotAxis, colorbar

def plot_dose_map(map_to_plot,
                  plot_title=None,
                  plot_contours=True,
                  levels=3,
                    **kwargs):

    #altitude_to_plot_in_km = altitude_to_plot_in_kft * 0.3048

    axis_no_colorbar, colorbar = create_single_dose_map_plot_plt(map_to_plot,
                                                     **kwargs)

    plt.title(plot_title)

    if plot_contours is True:
        plot_dose_map_contours(map_to_plot,levels=levels,**kwargs)

    return axis_no_colorbar, colorbar

def add_colorbar_to_plot(hue_range, palette, legend_label, scatterPlotAxis=None):
    norm = plt.Normalize(hue_range[0], hue_range[1])
    sm = plt.cm.ScalarMappable(cmap=palette, norm=norm)
    sm.set_array([])

    # Remove the legend and add a colorbar
    #scatterPlotAxis.get_legend().remove()
    if scatterPlotAxis is None:
        colorbar = plt.colorbar(sm,label=legend_label,shrink=0.4)
    else:
        colorbar = scatterPlotAxis.figure.colorbar(sm,label=legend_label,shrink=0.4)
    return colorbar

def create_single_dose_map_plotly(DF_to_use,
                              selected_altitude_in_km,
                              **kwargs):

    if selected_altitude_in_km is not None:
        DF_to_use = DF_to_use[round(DF_to_use["altitude (km)"],4) == selected_altitude_in_km]

    if len(DF_to_use) == 0:
        raise Exception("Error: specified altitude in kilometers did not match any of the altitudes in kilometers in the inputted DataFrame.")

    doseRateMap = px.scatter(DF_to_use, x="longitude",y="latitude",color="adose",
                            symbol_sequence=["square"],
                            range_y=[-90,90],
                            range_x=[0,360],
                            **kwargs)

    doseRateMap.update_traces(marker={'size': 10})
    doseRateMap.update_layout(yaxis=dict(scaleanchor="x", scaleratio=1, range=[-90,90]))
    doseRateMap.update_xaxes(range=[0,360])
    doseRateMap.update_yaxes(range=[-90,90])
    doseRateMap.update_layout(xaxis_scaleanchor="y")

    doseRateMap.update_layout(autosize=False,
                            width=800,
                            height=600)

    doseRateMap.show()

    return doseRateMap


import numpy as np
import pandas as pd
from pandarallel import pandarallel
pandarallel.initialize(progress_bar=True)
from atmosphericRadiationDoseAndFlux import doseAndFluxCalculator as DAFcalc
from scipy.interpolate import interp1d
from scipy.integrate import trapz
import ParticleRigidityCalculationTools as PRCT

from .data.NM64_responses import get_NM64_response_value_altitude, response_value_epns
from .AsymptoticDirectionProcessing import acquireWeightingFactors, get_apply_method
from .spectralCalculations.particleDistribution import particleDistribution

class singleParticleEngineInstance():

    def __init__(self, 
                 particle_distribution:particleDistribution, 
                 dfofAsymptoticDirections:pd.DataFrame, 
                 list_of_altitudes_in_km:list,
                 generate_NM_count_rates:bool):
        
        self.particle_distribution = particle_distribution
        self.list_of_altitudes_in_km = list_of_altitudes_in_km
        self.dfOfAllAsymptoticDirections = dfofAsymptoticDirections
        self.generate_NM_count_rates=generate_NM_count_rates
        self.rigiditySpectrumParamDict = {}
        self.pitchAngleDistributionParamDict = {}

    def getThePitchAngleDistribution(self):
        print("acquiring pitch angle distributions...")
        pitchAngleDistToUse = self.particle_distribution.momentum_distribution.pitch_angle_distribution
        try:
            print("IMF latitude set to",self.IMFlatitude)
            print("IMF longitude set to",self.IMFlongitude)
        except AttributeError:
            self.IMFlatitude = -1000.0 # distribution is probably isotropic: while a better solution is probably necessary, set to dummy variables for now.
            self.IMFlongitude = -1000.0
        pitchAngleDistToUse.setInterplanetaryMagFieldDirection(self.IMFlatitude,self.IMFlongitude)
        return pitchAngleDistToUse

    def getAsymptoticDirsAndRun(self):

        self.acquireDFofAllAsymptoticDirections()
        sortedOutputDoseRates = self.runOverSpecifiedAltitudes()

        return sortedOutputDoseRates

    def runOverSpecifiedAltitudes(self):
        
        #######################################################
        # assign each asymptotic direction a weighting factor in accordance with its pitch angle
        print("assigning pitch angle weighting factors...")
        df_with_weighting_factors_full_angles = acquireWeightingFactors(self.dfOfAllAsymptoticDirections,
                                                            self.particle_distribution)
        #######################################################

        # print("df_with_weighting_factors:")
        # print(df_with_weighting_factors.to_string())
        # import pickle as pkl
        # with open("weightingFactorsDF_2.pkl","wb") as weightingFactorsFile:
        #     pkl.dump(df_with_weighting_factors, weightingFactorsFile)

        #######################################################
        df_with_weighting_factors_full_angles.to_pickle("df_with_weighting_factors_full_angles.pkl")

        df_with_weighting_factors = get_mean_weighting_factors_for_multi_angle_magcos_runs(df_with_weighting_factors_full_angles)
        #######################################################

        #######################################################
        # calculate dose rates at each point based on spectra and weighting factors
        print("converting spectra and asymptotic directions to particle fluxes and dose rates...")

        sortedOutputDoseRates = self.calc_output_dose_flux(df_with_weighting_factors, 
                                                           self.list_of_altitudes_in_km, 
                                                           self.particle_distribution.particle_species.particleName)
        
        df_with_weighting_factors.to_pickle("weighting_factor_DF.pkl")
        sortedOutputDoseRates.weighting_factor_input_DF = df_with_weighting_factors
        
        print("output dose rates calculated successfully!")

        return sortedOutputDoseRates

    def calc_output_dose_flux(self, asymp_dir_DF_with_weighting_factors, list_of_altitudes_in_km, particle_name):

        spectrum_to_function_conversion_function = lambda row: interp1d(x=np.array(row["Rigidity"]),
                                                                        y=np.array(row["fullRigidityPitchWeightingFactor"]),
                                                                        bounds_error=False,
                                                                        fill_value=0.0)

        DFofSpectraForEachCoord = asymp_dir_DF_with_weighting_factors.groupby(["initialLatitude","initialLongitude"]).apply(spectrum_to_function_conversion_function)
        outputDoseRatesForAltitudeRange = get_apply_method(DFofSpectraForEachCoord)(lambda spectrum:DAFcalc.calculate_from_rigidity_spec(
                                                                                                inputRigidityDistributionFunctionGV=lambda x:float(spectrum(x)),
                                                                                                altitudesInkm=list_of_altitudes_in_km,
                                                                                                particleName = particle_name,
                                                                                                )
                                                                            )
        
        outputDoseRatesOnlyDF = pd.concat(outputDoseRatesForAltitudeRange.tolist(), ignore_index=True)

        index_list = [-2,-1,0,1,2,3,4,5,6,7,8]

        if self.generate_NM_count_rates == True:
            print("calculating neutron monitor count rates...")
            outputDoseRatesOnlyDF["NM64_cr_unnorm"] = self.get_neutron_monitor_count_rates(list_of_altitudes_in_km, particle_name, DFofSpectraForEachCoord)
            print("neutron monitor count rates successfully determined.")

            index_list = [-2,-1,0,1,2,3,4,5,6,7,8,9]
        
        outputDoseRatesOnlyDF_with_lats_and_longs = self.reinsert_original_lats_and_longs(outputDoseRatesOnlyDF,outputDoseRatesForAltitudeRange)

        sortedOutputDoseRates = outputDoseRatesOnlyDF_with_lats_and_longs  \
                                        .iloc[:,index_list] \
                                        .sort_values(["latitude","longitude","altitude (km)"], ignore_index=True)
        return sortedOutputDoseRates

    def get_neutron_monitor_count_rates(self, list_of_altitudes_in_km, particle_name, DFofSpectraForEachCoord):
        NM64_values_list = []

        for altitude_in_km in list_of_altitudes_in_km:
            NM64_vals_DF = get_apply_method(DFofSpectraForEachCoord)(lambda x:self.calculate_unnormed_NM64_cr(x,particle_name = particle_name, altitude_in_km=altitude_in_km)).values
            NM64_values_list.append(NM64_vals_DF)

        new_cr_unnorm_list = np.array(NM64_values_list).T.flatten()
        return new_cr_unnorm_list

    def calculate_unnormed_NM64_cr(self, interpolated_spectra_row, particle_name, altitude_in_km = 0.0):
        response_values = get_NM64_response_value_altitude(particle_name, response_value_epns, altitude_in_km)

        NM64_response_tabulated_2020_functional_epn = pd.DataFrame({"Energy_per_nucleon_GeV_per_n":response_value_epns,
                                                                        "Yield_per_m2_per_sr":response_values})

        if particle_name == "proton":

            NM64_response_tabulated_2020_functional = PRCT.convertParticleEnergySpecToRigiditySpec(NM64_response_tabulated_2020_functional_epn["Energy_per_nucleon_GeV_per_n"]*1000,
                                                    fluxInEnergyMeVform=NM64_response_tabulated_2020_functional_epn["Yield_per_m2_per_sr"],
                                                    particleMassAU=1,
                                                    particleChargeAU=1)
            
        elif particle_name == "alpha":

            NM64_response_tabulated_2020_functional = PRCT.convertParticleEnergySpecToRigiditySpec(NM64_response_tabulated_2020_functional_epn["Energy_per_nucleon_GeV_per_n"] * 1000 * 4,
                                                    fluxInEnergyMeVform=NM64_response_tabulated_2020_functional_epn["Yield_per_m2_per_sr"],
                                                    particleMassAU=4,
                                                    particleChargeAU=2)

        else:
            raise Exception("ERROR: particle name did not match either proton or alpha!")
        
        NM64_response_to_use = NM64_response_tabulated_2020_functional

        spectra_weighting_factors_across_NM = NM64_response_to_use["Rigidity"].apply(interpolated_spectra_row)
        
        full_NM64_spectra_weighting_factor = spectra_weighting_factors_across_NM * NM64_response_to_use["Rigidity distribution values"]
        output_unnormalised_NM64_cr = trapz(full_NM64_spectra_weighting_factor, NM64_response_to_use["Rigidity"])
        return output_unnormalised_NM64_cr

    def reinsert_original_lats_and_longs(self, new_doserates_DF:pd.DataFrame,original_data_frame:pd.Series):
        original_lats = original_data_frame.reset_index(level=[0,1])["initialLatitude"]
        original_longs = original_data_frame.reset_index(level=[0,1])["initialLongitude"]

        list_of_altitude_dfs = []
        for altitude_in_km in new_doserates_DF["altitude (km)"].unique():
            df_for_specific_altitude = new_doserates_DF[new_doserates_DF["altitude (km)"] == altitude_in_km].reset_index(drop=True).copy()
            df_for_specific_altitude["latitude"] = original_lats
            df_for_specific_altitude["longitude"] = original_longs
            list_of_altitude_dfs.append(df_for_specific_altitude)

        return pd.concat(list_of_altitude_dfs, ignore_index=True)
    
def get_max_weighting_factors_for_multi_angle_magcos_runs(multi_angle_DF:pd.DataFrame):

    max_weighting_factor_DF = multi_angle_DF[multi_angle_DF[["Rigidity","initialLatitude","initialLongitude","fullRigidityPitchWeightingFactor"]].groupby(["Rigidity","initialLatitude","initialLongitude"], group_keys=False).apply(lambda x:x == x.max())["fullRigidityPitchWeightingFactor"]]

    return max_weighting_factor_DF.reset_index(drop=True)

def get_mean_weighting_factors_for_multi_angle_magcos_runs(multi_angle_DF:pd.DataFrame):

    max_weighting_factor_DF = multi_angle_DF.groupby(["Rigidity","initialLatitude","initialLongitude"], group_keys=False).mean()

    return max_weighting_factor_DF.reset_index()